

#include "Capri.h"
#include "LibUtilityTable.h"


static int utilityIndex=0;
static mapAttribute attrTable;
static mapUtility utilTable;
static mapFlow flowTable;
static mapUnknownConst unknownConstTable;
static mapString defaultUtilityStringTable;
static mapString defaultConstraintStringTable;
static mapTree defaultUtilityTreeTable;
static mapTree defaultConstraintTreeTable;
static mapRange rangeTable;
static mapApp appTable;

static CRITICAL_SECTION tableLock;
#define AcquireLock {EnterCriticalSection(&tableLock);}
#define ReleaseLock {LeaveCriticalSection(&tableLock);}


void initUtilityTable()
{
	InitializeCriticalSection(&tableLock);
}

void closeUtilityTable()
{
	DeleteCriticalSection(&tableLock);
	attrTable.clear();
	utilTable.clear();
	flowTable.clear();
	unknownConstTable.clear();

	defaultUtilityStringTable.clear();
	defaultConstraintStringTable.clear();

	defaultUtilityTreeTable.clear();
	defaultConstraintTreeTable.clear();
}

void LockTable()    {AcquireLock;}
void UnlockTable()  {ReleaseLock;}

template<class T>
bool isMatched(T in1, T in2)
{
	bool matched;
	if (in1 == ANY)
	{
		matched = TRUE;
	}
	else
	{
		matched = (in1 == in2) ? TRUE : FALSE;
	}
	return matched;
}

INT getConfigurabilityType(char *conf)
{
	if (!strcmp(conf, "static"))
	{
		return CONFIGURABILITY_STATIC;
	}
	else if(!strcmp(conf, "dynamic"))
	{
		return CONFIGURABILITY_DYNAMIC;
	}
	else
	{
		return CONFIGURABILITY_STATIC;
	}
}

INT getConfigurability(const INT fid)
{
	FLOAT ret;
	ret = getAttribute(fid, CONFIGURABILITY);
	return (INT)ret; 
}

INT putConfigurability(const INT fid, INT configurability) 
{
	INT ret;
	ret = putAttribute(fid, CONFIGURABILITY, (float)configurability);
	return ret;
}

//remove all statically registered configurations
INT removeStaticConfigurations(vector<FlowID> *removedFlowId)
{
	int size;
	size = (int)attrTable.size();
	for (int i = 0; i < size; i++)
	{
		if ((int)attrTable[i][CONFIGURABILITY] == CONFIGURABILITY_STATIC)
		{
			//fprintf(stderr, "removing.. flow index %d with FlowID %d %d %d %d %d\n",i, flowTable[i].dstAddr, flowTable[i].dstPort, flowTable[i].protocol, flowTable[i].srcAddr, flowTable[i].srcPort);
			removedFlowId->push_back((FlowID)flowTable[i]);
			utilTable.erase(i);
			flowTable.erase(i);
			attrTable.erase(i);
			unknownConstTable.erase(i);
		}
		else{
			//fprintf(stderr, "not remove fid %d attrTable[i][CONFIGURABILITY] %d\n", i, (int)attrTable[i][CONFIGURABILITY]);
		}
	}
	
	defaultUtilityStringTable.clear();
	defaultConstraintStringTable.clear();
	
	defaultConstraintTreeTable.clear();
	defaultUtilityTreeTable.clear();

	return CAPRI_OK;
}

INT removePorts()
{
	rangeTable.clear();
	return CAPRI_OK;
}

int getTableIndex(const FlowID *Id)
{
	
	int ret = CAPRI_INDEX_NOT_FOUND;
	bool matched = FALSE;
	
	if (Id==NULL) return ret;

	// check Flow Table if Id is already in the table
	int size = utilTable.size();
	for (int i=0; i< size; i++) 
	{
		if(utilTable.count(i) && flowTable.count(i))
		{
			matched = isMatched(Id->protocol, flowTable[i].protocol);
			if (matched) matched &= isMatched(Id->srcAddr, flowTable[i].srcAddr);
			if (matched) matched &= isMatched(Id->dstAddr, flowTable[i].dstAddr);
			if (matched) matched &= isMatched(Id->srcPort, flowTable[i].srcPort);
			if (matched) matched &= isMatched(Id->dstPort, flowTable[i].dstPort);
			if (matched) 
			{
				ret = i;
				break;
			}
		}
	}
	//fprintf(stderr, "CAPRI_INDEX_NOT_FOUND\n");
	return ret;
}

int getAttributeIndex(const INT fid, string attribute)
{
	int ret; 

	static int ui = UNKNOWN_ATTRIBUTE;
	
	// new entry
	if (unknownConstTable[fid].find(attribute) == unknownConstTable[fid].end())
	{
		unknownConstTable[fid].insert(pair<string, int>(attribute, ui));
		ret = ui;
		ui++;
	}
	else
	{
		ret = unknownConstTable[fid][attribute];
	}
	return ret;
}


float getMin(const INT fid, int attr) 
{
	if(attrTable[fid].find(attr + MIN_OFFSET) != attrTable[fid].end())
	{
		return attrTable[fid][attr + MIN_OFFSET];
	}
	return 0;
}


float getMax(const INT fid, int attr) 
{
	if(attrTable[fid].find(attr + MAX_OFFSET) != attrTable[fid].end())
	{
		return attrTable[fid][attr + MAX_OFFSET];
	}
	return CAPRI_FAILED;
}

float getCurMax(const INT fid, int attr) 
{
	if(attrTable[fid].find(attr + CUR_MAX_OFFSET) != attrTable[fid].end())
	{
		return attrTable[fid][attr + CUR_MAX_OFFSET];
	}
	return CAPRI_FAILED;
}

bool checkAvailableConstraint(const INT fid, int attr)
{
	bool ret = TRUE;
	if(attrTable[fid].find(attr+MIN_OFFSET) == attrTable[fid].end() || 
	   attrTable[fid].find(attr+MAX_OFFSET) == attrTable[fid].end()	)
	{
		ret = FALSE;
	}
	return ret;
}

float getAttribute(const INT fid, int attr)
{
	float ret;
	ret = attrTable[fid][attr];
	return ret;
}

float getUtility(const INT fid, int utilIndex)
{
	float ret;
	ret = attrTable[fid][utilIndex];
	return ret;
}

float getAveragedUtility(const INT fid, int action, int start_index, int window_size)
{
	float ret, sumUtil=0, temp=0;
	int i, size;

	switch(action)
	{
		case AVERAGE_MOVING:
			size = 0;
			for (i = 0; i < window_size; i++)
			{
				temp = attrTable[fid][UTILITY_VALUE + getUtilityIndex(start_index, -i)];
				if (temp > 0)
				{
					sumUtil += temp;
					size++;
				}
			}
			ret = sumUtil/size;

		break;

		case AVERAGE_MEDION:

		break;
	}

	return ret;
}

int getUtilityIndex(int utilIndex, int shift)
{
	int ret;

	switch(shift)
	{
		case LAST_UTILITY:
			ret = utilIndex - 1;
			if(ret < 0) ret += MAX_UTILITY_BUFFER;
			else if(ret >= MAX_UTILITY_BUFFER) ret %= MAX_UTILITY_BUFFER;
		break;

		case AVERAGE_MOVING:
			ret = AVERAGE_INDEX;
		break;
		default:
			ret = utilIndex + shift;
			if(ret < 0) ret += MAX_UTILITY_BUFFER;
			else if(ret >= MAX_UTILITY_BUFFER) ret %= MAX_UTILITY_BUFFER;
		break;
	
	}
	
	return ret;
}

float getUtilityDiff(const INT fid, int utilIndex, int shift)
{
	float ret1, ret2, diff;
	
	ret1 = attrTable[fid][UTILITY_VALUE + utilIndex];
	ret2 = attrTable[fid][UTILITY_VALUE + getUtilityIndex(utilIndex, shift)];
	diff = fabs((float)ret1-ret2);
	return diff;
}

float getUtilityDiffFromAvg(const INT fid, int action, int utilIndex)
{
	float ret1, ret2, diff;

	ret1 = attrTable[fid][utilIndex];
	ret2 = getAveragedUtility(fid, action, utilIndex, MOVING_AVERAGE_SIZE);
	diff = fabs((float)ret1-ret2);

	return diff;
}

int setMax(const INT fid, int attr, float val)
{
	if(attrTable[fid].find(attr + MAX_OFFSET) != attrTable[fid].end())
	{
		attrTable[fid][attr + MAX_OFFSET] = val;
		return CAPRI_OK;
	}
	return CAPRI_FAILED;
}

int setCurMax(const INT fid, int attr, float val)
{
	if(attrTable[fid].find(attr + CUR_MAX_OFFSET) != attrTable[fid].end())
	{
		attrTable[fid][attr + CUR_MAX_OFFSET] = val;
		return CAPRI_OK;
	}
	return CAPRI_FAILED;
}

int setMin(const INT fid, int attr, float val)
{
	if(attrTable[fid].find(attr + MIN_OFFSET) != attrTable[fid].end())
	{
		attrTable[fid][attr + MIN_OFFSET] = val;
		return CAPRI_OK;
	}
	return CAPRI_FAILED;
}

int putAttribute(const INT fid, int attr, float val)
{
	if (attr == UTILITY_VALUE)
	{
		utilityIndex = getUtilityIndex(utilityIndex, 1); // increment utilityIndex by 1
		attrTable[fid][UTILITY_VALUE + utilityIndex] = val;
		attrTable[fid][UTILITY_INDEX] = (float)utilityIndex;

		attrTable[fid][UTILITY_AVERAGE] = getAveragedUtility(fid, AVERAGE_MOVING, utilityIndex, MOVING_AVERAGE_SIZE);
	}
	else 
	{
		attrTable[fid][attr] = val;
	}
	return CAPRI_OK;
}

int putNormalizedUtility(const INT fid, float val)
{
	return putAttribute(fid, NORMALIZED_UTILITY, val);
}

const FlowID *getFlowId(const INT fid)
{
	FlowID *ret;
	ret = (FlowID *)&flowTable[fid];
	return ret;
}

SOCKET getSocket(const INT aid)
{
	SOCKET ret;
	if (appTable.count(aid) > 0)
	{
		ret = (SOCKET)appTable[aid];
		return ret;
	}
	else
	{
		return NULL;
	}
}


bool fidIsAvailable(const INT fid)
{
	bool ret;
	ret = flowTable.find(fid) != flowTable.end() ? TRUE : FALSE;
	return ret;
}

INT putFlowId(const INT fid, const FlowID *id)
{
	FlowID *newFlow = new FlowID;

	memcpy(newFlow, id, sizeof(FlowID));
	flowTable.insert(pair<int,const FlowID>(fid, *newFlow));

	return 1;
}

INT putSocket(const INT aid, const SOCKET newApp)
{
	appTable.insert(pair<int, SOCKET>(aid, newApp));
	return 1;
}

INT putUtilityTree(const INT fid, utilityTree *utilTree)
{
	utilTable.insert(pair<int, utilityTree *>(fid, utilTree));
	return CAPRI_OK;
}

INT putDefaultConstraintTree(APP_TYPE appType, utilityTree *constTree) 
{
	defaultConstraintTreeTable.insert(pair<APP_TYPE, utilityTree *>(appType, constTree));
	return CAPRI_OK;
}
		
INT putDefaultUtilityTree(APP_TYPE appType, utilityTree *utilTree) 
{
	defaultUtilityTreeTable.insert(pair<APP_TYPE, utilityTree *>(appType, utilTree));
	return CAPRI_OK;
}

utilityTree *getDefaultUtilityTree(APP_TYPE appType)
{
	return defaultUtilityTreeTable[appType];
}

utilityTree *getDefaultConstraintTree(APP_TYPE appType)
{
	return defaultConstraintTreeTable[appType];
}

INT putDefaultUtilitySpec(INT appType, const char *utilSpec, const char * utilConst)
{
	defaultUtilityStringTable[appType].insert(0, utilSpec);
	defaultConstraintStringTable[appType].insert(0, utilConst);
	return CAPRI_OK;
}

INT putPortRange(APP_TYPE appType, INT start, INT end)
{
	rangeTable[start][end] = appType;
	return CAPRI_OK;
}

APP_TYPE getAppTypeByPortRange(INT port)
{
	mapRange::iterator p;
	map<int, APP_TYPE>::iterator p2;
	APP_TYPE ret = CAPRI_APP_DEFAULT;

	int start, end;
	
	for (p = rangeTable.begin(); p != rangeTable.end(); p++)
	{
		start = p->first;
		for (p2 = rangeTable[start].begin(); p2 != rangeTable[start].end(); p2++)
		{
			end = p2->first;
			if ((port >= start) && (port <= end))
			{
				ret = rangeTable[start][end];
			}
		}
	}
	return ret;
}

const char *getUtilitySpec(INT appType)
{
	return defaultUtilityStringTable[appType].c_str();
}

const char *getConstraintSpec(INT appType)
{
	return defaultConstraintStringTable[appType].c_str();
}

int getFreeIndex() 
{
	int i;
	int ret = CAPRI_INDEX_NOT_FOUND; 
	for (i=0 ; i < (int)utilTable.max_size(); i++)
	{
		if(utilTable.count(i) > 0) 
		{
		}
		else 
		{
			ret = i;
			break;
		}
	}
	
	return ret;
}

INT deregisterUtilityFromTable(const INT fid) 
{
	utilTable.erase(fid);
	flowTable.erase(fid);
	attrTable.erase(fid);
	unknownConstTable.erase(fid);

	return CAPRI_OK;
}

utilityTree * getUtilityTree(const INT fid)
{
	if (fid >= 0) return utilTable[fid];
	return NULL;
}

mapUtility *getUtilityTable()
{
	return &utilTable;
}

mapFlow *getFlowTable()
{
	return &flowTable;
}

void printUtilityTable()
{
	mapUtility::iterator it;
	int i=0;
	
	fprintf(stderr, "++++++ printUtilityTable ++++++\n");

	for (it=utilTable.begin(); it!=utilTable.end(); it++)
	{
		fprintf(stderr, "count =%d   -- %f => %f\n", utilTable.count(i++), (*it).first, (*it).second);
	}
}

int getUtilityTableSize()
{
	return utilTable.size();
}

